@@ -1,6 +1,7 @@
 import collections
 from copy import deepcopy
 from enum import Enum
+import math
 import os
 
 import requests
@@ -9,6 +10,51 @@ from .base import GenericAPI, ERROR_THRESHOLD
 from .. import utils
 
 
+class UploadProgressHandler:
+    def __init__(self, chunk_size, total_size, callback):
+        if not isinstance(chunk_size, int) or chunk_size <= 0:
+            raise ValueError("chunk_size must be a positive integer")
+            
+        if not isinstance(total_size, int) or total_size <= 0:
+            raise ValueError("total_size must be a positive integer")
+        
+        if not callable(callback):
+            raise ValueError("callback must be callable")
+            
+        # TODO:  Callback must accept an integer (number of chunks already added).
+            
+        self.chunk_size = chunk_size
+        self.total_size = total_size
+        self.callback = callback
+        
+        # Invoke callback when current size is >= threshold.
+        self.current_size = 0
+        self.callback_threshold = chunk_size
+        
+    def add(self, increment):
+        # increment must be a positive integer, though in the interest of performance
+        # that won't be checked.
+        if not increment:
+            return
+
+        self.current_size += increment
+        
+        if self.current_size >= self.total_size:
+            # NOTE:  Do not use self.current_size as the numerator.
+            #        It may be larger than self.total_size, for reasons unknown.
+            self.callback(math.ceil(self.total_size / self.chunk_size))
+            return
+        
+        if self.current_size >= self.callback_threshold:
+            self.callback(math.floor(self.current_size / self.chunk_size))
+
+            while self.callback_threshold <= self.current_size:
+                self.callback_threshold = min(
+                    self.callback_threshold + self.chunk_size,
+                    self.total_size
+                )
+
+
 class FilesAPI(GenericAPI):
     """
     Files API switchboard.
@@ -472,8 +518,10 @@ class FilesAPI(GenericAPI):
         else:
             return r.json()
 
-    def upload(self, local_path, space=None, space_id=None, path=None,
-            metadata=None):
+    def upload(self, local_path,
+               space=None, space_id=None, path=None,
+               metadata=None,
+               progress_callback=None, n_chunks=100,):
         """
         Upload a file to HyperThought.
 
@@ -494,6 +542,12 @@ class FilesAPI(GenericAPI):
                 ',uuid,uuid,uuid,'
         metadata : dict-like or None
             Metadata for the file.
+        progress_callback : callable (int -> None) or None
+            A callback for handling upload progress.  Will be called each time
+            a given number of bytes (chunk) is uploaded.
+        n_chunks : int
+            The number of chunks to be handled by progress_callback.
+            Will be ignored if progress_callback is None.
 
         Returns
         -------
@@ -511,6 +565,13 @@ class FilesAPI(GenericAPI):
             
         metadata = self._validate_metadata(metadata)
 
+        # TODO:  Move this into a validation function.
+        if progress_callback is not None and not callable(progress_callback):
+            raise ValueError("progress_callback must be a callable or None")
+
+        if not isinstance(n_chunks, int):
+            raise ValueError("n_chunks must be an int")
+
         # Get file name and size using the local path.
         active_local_path = utils.get_active_path(local_path)
         name = active_local_path.split(os.path.sep)[-1]
@@ -527,7 +588,12 @@ class FilesAPI(GenericAPI):
         )
 
         # Use the url to upload the file.
-        self._upload_using_url(url, active_local_path)
+        self._upload_using_url(
+            url,
+            active_local_path,
+            progress_callback,
+            n_chunks,
+        )
 
         # Move the file from the temporary to the permanent file collection.
         file_name = self._temp_to_perm(file_id)
@@ -664,7 +730,7 @@ class FilesAPI(GenericAPI):
         if 'ftype' not in document['content']:
             return False
 
-        return document['content']['ftype'] == FOLDER_TYPE
+        return document['content']['ftype'] == utils.FOLDER_TYPE
 
     def _get_upload_url(self, name, size, space='user', space_id=None,
                         path=None, metadata=None):
@@ -737,7 +803,8 @@ class FilesAPI(GenericAPI):
 
         return url, file_id
 
-    def _upload_using_url(self, upload_url, local_path):
+    def _upload_using_url(self, upload_url, local_path,
+                          progress_callback, n_chunks):
         """
         Use a url to upload a file.
 
@@ -749,36 +816,59 @@ class FilesAPI(GenericAPI):
             The url to which the file should be uploaded.
         local_path : str
             The local path to the file to be uploaded.
+        progress_callback : callable(int -> None) or None
+            A callable to provide progress on upload status.
+        n_chunks : int
+            The number of chunks to be handled by progress_callback.
+            Will be ignored if progress_callback is None.
         """
         upload_url = self._validate_url(upload_url)
         local_path = self._validate_local_path(local_path)
+        
+        file_size = os.path.getsize(local_path)
+        file_handle = open(local_path, 'rb')
 
-        with open(local_path, 'rb') as file_handle:
-            kwargs = {
-                'url': upload_url,
-                'data': file_handle,
-                'verify': False,
-                'stream': True,
-                'headers': {},
-            }
+        if progress_callback is not None:
+            chunk_size = math.ceil(file_size / n_chunks)
+            progress_handler = UploadProgressHandler(
+                chunk_size=chunk_size,
+                total_size=file_size,
+                callback=progress_callback,
+            )
+            original_read = file_handle.read
+            
+            def new_read(size):
+                progress_handler.add(size)
+                return original_read(size)
+                
+            file_handle.read = new_read
+        
+        kwargs = {
+            'url': upload_url,
+            'data': file_handle,
+            'verify': False,
+            'stream': True,
+            'headers': {},
+        }
 
-            if self.get_backend() == 'default':
-                kwargs['headers'].update(self._auth.get_headers())
-                # Content-Disposition (with file name) is required by Django 2.2.
-                file_name = local_path.strip(utils.PATH_SEP).split(utils.PATH_SEP)[-1]
-                kwargs['headers']['Content-Disposition'] = f"inline;filename={file_name}"
+        if self.get_backend() == 'default':
+            kwargs['headers'].update(self._auth.get_headers())
+            # Content-Disposition (with file name) is required by Django 2.2.
+            file_name = local_path.strip(utils.PATH_SEP).split(utils.PATH_SEP)[-1]
+            kwargs['headers']['Content-Disposition'] = f"inline;filename={file_name}"
+            kwargs['cookies'] = self._auth.get_cookies()
+        else:
+            # TODO:  Why can't this be removed?
+            kwargs['headers'].update({
+                'Content-Type': 'application/octet-stream'
+            })
 
-                kwargs['cookies'] = self._auth.get_cookies()
-            else:
-                # TODO:  Why can't this be removed?
-                kwargs['headers'].update({
-                    'Content-Type': 'application/octet-stream'
-                })
-
-            r = requests.put(**kwargs)
+        r = requests.put(**kwargs)
 
         if r.status_code >= ERROR_THRESHOLD:
             self._report_api_error(response=r)
+            
+        file_handle.close()
 
     def _temp_to_perm(self, file_id):
         """
